	
;simple non-updating scroller (512x512) with automatic wrap-around

	cpu 68000
	supmode on

	include "reg.inc"

;RAM
	org $100000
vblankflag:	ds.w 1	;signal for main/vblank processing
xscroll:	ds.w 1	;current xscroll position
yscroll:	ds.w 1	;current yscroll position

;ROM header
	org $0064
	dc.l VBLANK	;IRQ handler

	org $0100
	dc.b "NEO-GEO",$00

	org $0108
	dc.w $017	;Sengoku

	org $0122
	jmp Start	;entry

	org $0114
	dc.w $0100	;logo flag, don't show it just go straight to the entry point

	org $0182
	dc.l Code	;code pointer
Code:
	dc.l $76004A6D,$0A146600,$003C206D,$0A043E2D
	dc.l $0A0813C0,$00300001,$32100C01,$00FF671A
	dc.l $30280002,$B02D0ACE,$66103028,$0004B02D
	dc.l $0ACF6606,$B22D0AD0,$67085088,$51CFFFD4
	dc.l $36074E75,$206D0A04,$3E2D0A08,$3210E049
	dc.l $0C0100FF,$671A3010,$B02D0ACE,$66123028
	dc.l $0002E048,$B02D0ACF,$6606B22D,$0AD06708
	dc.l $588851CF,$FFD83607
	dc.w $4e75

;-----

;code

XINIT	SET 0	;initial xscroll position
YINIT	SET 0	;initial yscroll position
WIDTH	SET 32	;width of map in tiles (512 pixel / 16)
HEIGHT	SET 32	;height of map in tiles (512 pixel / 16)

Start:
	lea	VRAM_RW,A6	;A6 always has $3C0002
	jsr	ClearFIX	;no fix tiles on screen
	jsr	SetPalette
	jsr	SetSCB1
	jsr	SetSCB2
	jsr	SetSCB3
	move.w	#XINIT,xscroll		;set initial xscroll
	move.w	#YINIT,yscroll		;set initial yscroll
	and	#$F0FF,SR	;allow interrupts now
.loop
	jsr	ControllerScroll
	st	vblankflag	;signal that main loop is processed
.spin
	tst.b	vblankflag	;wait until vblank happens
	bne	.spin
	bra	.loop

;---

ControllerScroll:
	move.w	xscroll,D0		;load scrolls to potentially modify
	move.w	yscroll,D1
	move.b	BIOS_P1CURRENT,D2	;load controller data to test
.testLeft
	btst	#CNT_LE,D2	;scroll to the left?
	beq	.testRight
.left
	subq.w	#2,D0		;xscroll--
	bra	.testUp
.testRight
	btst	#CNT_RI,D2	;scroll to the right?
	beq	.testUp
.right
	addq.w	#2,D0		;xscroll++
.testUp
	btst	#CNT_UP,D2	;scroll up?
	beq	.testDown
.up
	subq.w	#2,D1		;yscroll--
	bra	.storeScroll
.testDown
	btst	#CNT_DN,D2	;scroll down?
	beq	.storeScroll
.down
	addq.w	#2,D1		;yscroll++
.storeScroll
	move.w	D0,xscroll	;set scrolls for next vblank
	move.w	D1,yscroll
	rts

;---

ClearFIX:
	move.w	#FIXMAP,-2(A6)	;writing to FIX layer
	move.w	#1,2(A6)
	move.w	#$00FF,D0	;transparent tile
	move.w	#$500-1,D1	;clear entire map
.loop
	move.w	D0,(A6)
	dbra	D1,.loop
	rts

;---

SetPalette:
	clr.w	$400000		;black blanking color
	lea	.palette,A0	;pointer to palette in ROM
	move.w	(A0)+,$401FFE	;backdrop color is color 0 from palette
	lea	$400002,A1	;palette RAM target
	moveq	#15-1,D0	;15 colors remaining to copy
.loop
	move.w	(A0)+,(A1)+
	dbra	D0,.loop
	rts
.palette
	BINCLUDE "pal.bin"

;---

SetSCB1:
	move.w	#SCB1+$40,-2(A6);write to SCB1 starting from sprite 1, NOT sprite 0
	move.w	#1,2(A6)	
	moveq	#32-1,D1	;32 sprites to update
.loopOuter
	moveq	#WIDTH-1,D2
	sub.w	D1,D2			;outer loop counter is effectively the column counter
	lsl.w	#2,D2			;*4, 4 bytes per tile in map file
	lea	.map(PC,D2.w),A0	;A0: address to column in map file to upload to SCB1
	moveq	#32-1,D0		;32 pairs of words to write for each sprite tile
.loopInner
	move.w	(A0)+,(A6)		;tile low word (+2 from post increment compensated for below)
	move.w	(A0),(A6)		;tile high word
	lea	WIDTH*4-2(A0),A0	;A0 += (row size - 2), move to the next row in map file
	dbra	D0,.loopInner
.endInner
	dbra	D1,.loopOuter
.endOuter
	rts
.map
	BINCLUDE "map.bin"

;---

SetSCB2:
	move.w	#SCB2+1,-2(A6)	;writing to SCB2
	move.w	#1,2(A6)
	move.w	#$0FFF,D0	;X zoom max, Y zoom max
	move.w	#32-1,D1	;32 sprites are used in total
.loop
	move.w	D0,(A6)
	dbra	D1,.loop
	rts

;---

SetSCB3:
	move.w	#SCB3+2,-2(A6)		;writing to SCB3 sprite 2 onwards (sprite 1 SCB3 is set in UpdateVRAM)
	move.w	#1,2(A6)
	move.w	#$0040,D0		;set chained sprites chain bit, automatically inherits other properties from previous sprite
	moveq	#31-1,D1		;31 sprites are chained to master sprite
.loop
	move.w	D0,(A6)
	dbra	D1,.loop
	rts

;---

VBLANK:
	move.b	D0,$300001	;watchdog
	tst.b	$10FD80		;BIOS wants vblank?
	bmi	.regular
	jmp	$C00438
.regular
	move.w	#4,REG_IRQACK
	tst.b	vblankflag	;don't do anything if main routine hasn't finished yet
	beq	.return
.process
	movem.l	D0-D7/A0-A6,-(A7)		
	jsr	UpdateVRAM	;make VRAM changes needed for scrolling
	jsr	BIOSF_VBL	;read controller etc.
	movem.l	(A7)+,D0-D7/A0-A6
	clr.b	vblankflag	;signal that vblank is processed
.return
	rte

;---

;simple routine that applies VRAM values for scroll and nothing else

UpdateVRAM:
;SCB3
	move.w	#SCB3+1,-2(A6)	;writing to ypos of master sprite
	move.w	yscroll,D0
	add.w	#$01F0,D0	;screen Y coord based from $01F0
	asl.w	#7,D0		;ypos is in upper 9 bits
	or.w	#$0020,d0	;sprite size is $20 tiles tall
	move.w	D0,(A6)
;SCB4
	move.w	#SCB4+1,-2(A6)	;writing to xpos of master sprite
	move.w	xscroll,D0
	neg.w	D0
	asl.w	#7,D0		;xpos is in upper 9 bits
	move.w	D0,(A6)
	rts