	
;updating scroller (1024x1024 using 512x512 sprite space) with bounds checking

	cpu 68000
	supmode on

	include "reg.inc"

;RAM
	org $100000
vblankflag:	ds.w 1		;signal for main/vblank processing
xscroll:	ds.w 1		;current xscroll position
yscroll:	ds.w 1		;current yscroll position
prevxscroll:	ds.w 1		;previous frame xscroll position
prevyscroll:	ds.w 1		;previous frame yscroll position
xdir:		ds.b 1		;direction of xscroll this frame
ydir:		ds.b 1		;direction of yscroll this frame
vramlist:	ds.w 512	;list of addresses,increments and data to upload to VRAM
vramptr:	ds.d 1		;current pointer into vramlist, reset every frame

;vramlist has a number of entries formatted as such:
;word 0: number of words in this entry
;word 1: VRAM addr
;word 2: VRAM mod
;word n: data words

;ROM header
	org $0064
	dc.l VBLANK	;IRQ handler

	org $0100
	dc.b "NEO-GEO",$00

	org $0108
	dc.w $017	;Sengoku

	org $0122
	jmp Start	;entry

	org $0114
	dc.w $0100	;logo flag, don't show it just go straight to the entry point

	org $0182
	dc.l Code	;code pointer
Code:
	dc.l $76004A6D,$0A146600,$003C206D,$0A043E2D
	dc.l $0A0813C0,$00300001,$32100C01,$00FF671A
	dc.l $30280002,$B02D0ACE,$66103028,$0004B02D
	dc.l $0ACF6606,$B22D0AD0,$67085088,$51CFFFD4
	dc.l $36074E75,$206D0A04,$3E2D0A08,$3210E049
	dc.l $0C0100FF,$671A3010,$B02D0ACE,$66123028
	dc.l $0002E048,$B02D0ACF,$6606B22D,$0AD06708
	dc.l $588851CF,$FFD83607
	dc.w $4e75

;-----

;code

XINIT		SET 0	;initial xscroll position
YINIT		SET 0	;initial yscroll position
WIDTH		SET 64	;width of map in tiles
HEIGHT		SET 64	;height of map in tiles

Start:
	lea	VRAM_RW,A6	;A6 always has $3C0002
	jsr	ClearFIX	;no fix tiles on screen
	jsr	SetPalette	;setup palette
	jsr	SetSCB2
	jsr	SetSCB3
	move.w	#XINIT,xscroll		;set initial xscroll
	move.w	#YINIT,yscroll		;set initial yscroll
	move.l	#vramlist,vramptr	;initialise pointer to vramlist
	jsr	InitSCB1		;setup initial SCB1 graphics (first things you see) based on these scroll values
	and	#$F0FF,SR		;allow interrupts now
.loop
	move.w	xscroll,prevxscroll	;save copy of scrolls before modification
	move.w	yscroll,prevyscroll
	jsr	ControllerScroll	;scroll background
	jsr	ClampScroll		;make sure scroll stays within bounds
	jsr	CheckUpdate		;see if sprite VRAM has to be updated as a result of scroll changes
	jsr	BufferSCB3		;yscroll,ysize,chain
	jsr	BufferSCB4		;xscroll
	st	vblankflag		;signal that main loop is processed
.spin
	tst.b	vblankflag	;wait until vblank happens
	bne	.spin
	bra	.loop

Map:
	BINCLUDE "map.bin"

;---

;scrolls background according to your controller input

ControllerScroll:
	move.w	xscroll,D0		;load scrolls to potentially modify
	move.w	yscroll,D1
	move.b	BIOS_P1CURRENT,D2	;load controller data to test
.testLeft
	btst	#CNT_LE,D2	;scroll to the left?
	beq	.testRight
.left
	subq.w	#2,D0		;xscroll--
	bra	.testUp
.testRight
	btst	#CNT_RI,D2	;scroll to the right?
	beq	.testUp
.right
	addq.w	#2,D0		;xscroll++
.testUp
	btst	#CNT_UP,D2	;scroll up?
	beq	.testDown
.up
	subq.w	#2,D1		;yscroll--
	bra	.storeScroll
.testDown
	btst	#CNT_DN,D2	;scroll down?
	beq	.storeScroll
.down
	addq.w	#2,D1		;yscroll++
.storeScroll
	move.w	D0,xscroll	;set scrolls for next vblank
	move.w	D1,yscroll
	rts

;---

;keeps xscroll and yscroll within boundaries of the map

ClampScroll:
	move.w	xscroll,D0	;test xscroll
	tst.w	D0		;xscroll can't go negative
	bpl	.testRight
	clr.w	xscroll		;xscroll was beyond left boundary
	bra	.testY
.testRight
	move.w	#WIDTH*16-320,D1
	cmp.w	D1,D0		;xscroll can't go beyond the right boundary, taking into account the 320px screen width
	bls	.testY
	move.w	D1,xscroll	;xscroll was beyond right boundary
.testY
	move.w	yscroll,D0	;test yscroll
	tst.w	D0		;yscroll can't go negative
	bpl	.testBottom
	clr.w	yscroll		;yscroll was beyond top boundary
	bra	.return
.testBottom
	move.w	#HEIGHT*16-224,D1
	cmp.w	D1,D0		;yscroll can't go beyond the bottom boundary, taking into account the 224px screen height
	bls	.return
	move.w	D1,yscroll	;yscroll was beyond bottom boundary
.return
	rts

;---

;checks to see if there is a VRAM update for scrolling necessary

CheckUpdate:
	move.w	xscroll,D0	;load xscroll for testing
	move.w	prevxscroll,D1
	eor.w	D0,D1		;upper 12 bits will be zero if there is no difference
	and.w	#$FFF0,D1	;check if the scroll crossed a 16px boundary
	beq	.testY		;if there's no difference, then no update is needed
.xUpdatedNeeded
	cmp.w	prevxscroll,D0	;which direction was there scroll in?
	shi	xdir		;store direction for later reference (00 if scrolling left, FF if scrolling right)
	jsr	BufferSCB1X	;SCB1 VRAM must be updated now
.testY
	move.w	yscroll,D0	;load yscroll for testing
	move.w	prevyscroll,D1
	eor.w	D0,D1		;same test as above
	and.w	#$FFF0,D1
	beq	.return
.yUpdateNeeded
	cmp.w	prevyscroll,D0
	shi	ydir
	jsr	BufferSCB1Y	;SCB1 VRAM must be updated now
.return
	rts

;---

;buffers (and uploads) an entire screen of SCB1 VRAM

COLUMNS	SET 21	;update this many columns on screen (must fill the entire screen to look correct)

InitSCB1:
	move.w	xscroll,D0	;base xscroll (left of screen)
	lsr.w	#4,D0		;/16
	move.w	yscroll,D1	;base yscroll (top of screen)
	lsr.w	#4,D1		;/16
	moveq	#COLUMNS-1,D2	;update this many columns
.loop
	jsr	BufferSCB1Column	;prepare a column in the buffer
	jsr	UpdateVRAM		;upload the column immediately
	addq.w	#1,D0			;move 1 column (1 sprite,16pixels) towards the right
	dbra	D2,.loop
	rts

;---

;prepares SCB1 VRAM for scroll on X axis based on direction of scroll

LEFT	SET 0	;when scrolling left, update the column that is at left of screen
RIGHT	SET 20	;when scrolling right, update the column 20 columns to the left of screen

BufferSCB1X:
	move.w	xscroll,D0	;xscroll required to figure out what column of tiles to read from
	lsr.w	#4,D0		;/16, 16 pixels per sprite column
	move.w	#LEFT,D1	;offset value if scrolling left
	tst.b	xdir
	beq	+
	move.w	#RIGHT,D1	;offset value if scrolling right
+	add.w	D1,D0		;D0: xscroll value to start loading from
	move.w	yscroll,D1
	lsr.w	#4,D1		;D1: yscroll value to start loading from
	jsr	BufferSCB1Column
	rts

;writes column of SCB1 data to vramlist

;IN:
;D0.w: source xscroll (/16)
;D1.w: source yscroll (/16)

BUFFERCOUNT	SET 15	;15 sprite tiles to update in a column, enough to span from top to bottom of the screen

BufferSCB1Column:
	movem.l	D0-D3/A0/A1,-(SP)

	move.w	D1,D2		;copy yscroll to D2 for testing of wrap-around
	and.w	#$001F,D2	;mod 32, 32 sprite tiles per sprite column
	add.w	#BUFFERCOUNT,D2	;add number of tiles to buffer to see if it fits without wrap-around
	sub.w	#$0020,D2	;D2: number of tiles that wrapped-around and need to be handled in a separate write (0-14 tiles)
	bpl	+
	clr.w	D2		;underflow means there can't possibly be wrap-around
+	moveq	#BUFFERCOUNT,D3
	sub.w	D2,D3		;D3: number of tiles that didn't wrap around (1 tile minimum)

	jsr	GetIndexes	;get starting sprite tile in map and SCB1 VRAM
	move.w	D1,D4		;D4: saved copy of VRAM address possibly used in wrap-around upload
	lea	Map,A0
	add.l	D0,A0		;A0: starting address of sprite tile in map
	move.l	vramptr,A1	;A1: pointer to write vramlist
	move.w	D3,D0		;number of non-wrap-around tiles
	add.w	D0,D0		;*2, 2 words per sprite tile
	move.w	D0,(A1)+	;size = (D3) pairs of words to write
	move.w	D1,(A1)+	;vram addr = starting VRAM address in SCB1
	move.w	#1,(A1)+	;vram mod = 1, linear writes
	subq.w	#1,D3
.loop
	move.l	(A0),(A1)+	;one sprite tile written
	lea	WIDTH*4(A0),A0	;advance to next row in map (*4 since there are 4 bytes per tile)
	dbra	D3,.loop	;repeat for every non-wrap-around tile

	tst.w	D2		;are there any wrap-around tiles to handle?
	beq	.skip2		;if there are none to handle,don't bother buffering anything
.wrap
	move.w	D2,D0		;number of wrap-around tiles
	add.w	D0,D0		;*2, 2 words per sprite tile
	move.w	D0,(A1)+	;size = (D2) pairs of words to write
	and.w	#~$1F<<1,D4	;reset sprite tile # to write to in VRAM, effectively makes the target tile wrap around from bottom of sprite to top of sprite
	move.w	D4,(A1)+	;vram addr = starting VRAM address (at top of sprite) in SCB1
	move.w	#1,(A1)+	;vram mod = 1, linear writes
	subq.w	#1,D2
.loop2
	move.l	(A0),(A1)+	;one sprite tile written
	lea	WIDTH*4(A0),A0	;advance to next row in map
	dbra	D2,.loop2
.skip2
	move.l	A1,vramptr	;store updated vramptr
	movem.l	(SP)+,D0-D3/A0/A1
	rts

;---

TOP	SET 0	;when scrolling up, update the column at top of screen
BOTTOM	SET 14	;when scrolling down, update the column that is 14 tiles from top of screen

BufferSCB1Y:
	move.w	yscroll,D1	;yscroll required to figure out what row of tiles to read from
	lsr.w	#4,D1		;/16, 16 pixels per sprite column
	move.w	#TOP,D0		;offset value if scrolling up
	tst.b	ydir
	beq	+
	move.w	#BOTTOM,D0	;offset value if scrolling down
+	add.w	D0,D1		;D1: ypos to start loading from
	move.w	xscroll,D0
	lsr.w	#4,D0		;D0: xpos to start loading from
	jsr	BufferSCB1Row
	rts

;writes row of SCB1 data to vramlist

;IN:
;D0.w: source xscroll (/16)
;D1.w: source yscroll (/16)

BUFFERCOUNT	SET 21	;21 sprite tiles to update in a row, enough to span from left to right of the screen

BufferSCB1Row:
	movem.l	D0-D3/A0-A2,-(SP)

	move.w	D0,D2		;copy xscroll to D2 for testing of wrap-around
	and.w	#$001F,D2	;mod 32, 32 sprites total
	add.w	#BUFFERCOUNT,D2	;add number of tiles to buffer to see if it fits without wrap-around
	sub.w	#$0020,D2	;D2: number of tiles that wrapped-around and need to be handled in a separate write (0-20 tiles)
	bpl	+
	clr.w	D2		;underflow means there can't possibly be wrap-around
+	moveq	#BUFFERCOUNT,D3
	sub.w	D2,D3		;D3: number of tiles that didn't wrap around (1 tile minimum)

	jsr	GetIndexes	;get starting sprite tile in map and SCB1 VRAM
	move.w	D1,D4		;D4: saved copy of VRAM address
	lea	Map,A0
	add.l	D0,A0		;A0: starting address of row in map file

	move.l	vramptr,A1	;A1: pointer to write vramlist *low word only*
	move.w	D3,(A1)+	;size = (D3) pairs of words to write
	move.w	D1,(A1)+	;vram addr = starting vram address in SCB1 for this sprite tile
	move.w	#$0040,(A1)+	;vram mod = $40, number of words separating sprite columns

	move.w	D3,D0		;need size...
	add.w	D0,D0		;...*2
	lea	(A1,D0.w),A2	;A2: pointer to write vramlist *high word only*
	move.w	D3,(A2)+	;size = same as above
	move.w	D4,D0		;restore previous VRAM address
	addq.w	#1,D0		;+1 to write the upper word
	move.w	D0,(A2)+	;vram addr = same as above, but +1 to write upper word
	move.w	#$0040,(A2)+	;vram mod = same as above
	subq.w	#1,D3
.loop
	move.w	(A0)+,(A1)+	;sprite low word written
	move.w	(A0)+,(A2)+	;sprite high word written
	dbra	D3,.loop

	tst.w	D2		;are there any wrap-around tiles to handle?
	beq	.skip2
.wrap
	move.l	A2,A1		;for sake of consistency A1 will be the *low word only* list again
	move.w	D2,(A1)+	;size = (D2) pairs of words to write
	and.w	#~($1FF<<6),D4	;reset sprite # to write to in VRAM
	or.w	#$01<<6,D4	;set sprite # to sprite 1, effectively makes the target tile wrap around from left of sprites to right of sprites
	move.w	D4,(A1)+	;vram addr = starting VRAM address (at left of sprite) in SCB1
	move.w	#$0040,(A1)+	;vram mod = same as above

	move.w	D2,D0		;need size...
	add.w	D0,D0		;...*2
	lea	(A1,D0.w),A2	;A2: pointer to write vramlist *high word only*
	move.w	D2,(A2)+	;size = same as above
	move.w	D4,D0		;restore previous VRAM address
	addq.w	#1,D0		;+1 to write the upper word
	move.w	D0,(A2)+	;vram addr = same as above, but +1 to write upper word
	move.w	#$0040,(A2)+	;vram mod = same as above
	subq.w	#1,D2
.loop2
	move.w	(A0)+,(A1)+	;sprite low word written
	move.w	(A0)+,(A2)+	;sprite high word written
	dbra	D2,.loop2
.skip2
	move.l	A2,vramptr	;store updated vramptr
	movem.l	(SP)+,D0-D3/A0-A2
	rts

;---

;prepares SCB3 VRAM buffer contents

BufferSCB3:
	move.l	vramptr,A0	;A0: pointer to write vramlist
	move.w	#1,(A0)+	;1 word to write only
	move.w	#SCB3+1,(A0)+	;target sprite 1 in VRAM
	clr.w	(A0)+		;VRAM mod doesn't matter
	move.w	yscroll,D0
	add.w	#$01F0,D0	;screen Y coord based from $01F0
	asl.w	#7,D0		;set ypos in upper 9 bits
	or.w	#$0020,D0	;set size to $20 tiles with chain bit = 0, since this is the master sprite
	move.w	D0,(A0)+	;read current VRAM word
	move.l	A0,vramptr	;store updated vramptr
	rts

;---

;prepares SCB4 VRAM buffer contents

BufferSCB4:
	move.l	vramptr,A0	;A0: pointer to write vramlist
	move.w	#1,(A0)+	;1 word to write only
	move.w	#SCB4+1,(A0)+	;target sprite 1 in VRAM
	clr.w	(A0)+		;VRAM mod doesn't matter
	move.w	xscroll,D0
	neg.w	D0		;sprite scrolling to the left appears on screen as scroll to the right
	asl.w	#7,D0		;xpos is in upper 9 bits
	move.w	D0,(A0)+
	move.l	A0,vramptr	;store updated vramptr
	rts

;---

;returns indexes into the map data and VRAM

;IN:
;D0.w: xscroll	(/16)
;D1.w: yscroll	(/16)
;OUT:
;D0.l: map index
;D1.w: VRAM

GetIndexes:
	movem.w	D2/D3,-(SP)
	move.w	D0,D2		;save xpos in D2
	move.w	D1,D3		;save ypos in D3
;map index
	lsl.w	#2,D0		;xpos *4, gets index according to column
	mulu	#WIDTH*4,D1	;ypos *(row size), gets index according to row
	or.w	D0,D1		;combine column / row index bits
	move.l	D1,D0		;D0: map index
;VRAM index
	and.w	#$001F,D2	;clip xpos to 32 sprite range (32 sprites total used)
	addq.w	#1,D2		;start from sprite 1, not sprite 0
	lsl.w	#6,D2		;xpos *64, 64 words per sprite in SCB1
	and.w	#$001F,D3	;clip ypos to 32 sprite tile range (32 sprite tiles per sprite)
	add.w	D3,D3		;ypos *2, 2 words per sprite tile in SCB1
	or.w	D3,D2		;combine VRAM bits
	move.w	D2,D1		;D1: VRAM
	movem.w	(SP)+,D2/D3
	rts

;---

ClearFIX:
	move.w	#FIXMAP,-2(A6)	;writing to FIX layer
	move.w	#1,2(A6)
	move.w	#$00FF,D0	;transparent tile
	move.w	#$500-1,D1	;clear entire map
.loop
	move.w	D0,(A6)
	dbra	D1,.loop
	rts

;---

SetPalette:
	clr.w	$400000		;black blanking color
	lea	.palette,A0	;pointer to palette in ROM
	move.w	(A0)+,$401FFE	;backdrop color is color 0 from palette
	lea	$400002,A1	;palette RAM target
	moveq	#15-1,D0	;15 colors remaining to copy
.loop
	move.w	(A0)+,(A1)+
	dbra	D0,.loop
	rts
.palette
	BINCLUDE "pal.bin"

;---

SetSCB2:
	move.w	#SCB2+1,-2(A6)	;writing to SCB2
	move.w	#1,2(A6)
	move.w	#$0FFF,D0	;X zoom max, Y zoom max
	move.w	#32-1,D1	;32 sprites are used in total
.loop
	move.w	D0,(A6)
	dbra	D1,.loop
	rts

;---

SetSCB3:
	move.w	#SCB3+2,-2(A6)	;writing to SCB3 starting from sprite 2 (sprite 1 is the master and is set by updates in vramlist)
	move.w	#$0040,D0	;chain bit set for all successive sprites
	move.w	#31-1,D1	;31 sprites are chained to master sprite
.loop
	move.w	D0,(A6)
	dbra	D1,.loop
	rts

;---

VBLANK:
	move.b	D0,$300001	;watchdog
	tst.b	$10FD80		;BIOS wants vblank?
	bmi	.regular
	jmp	$C00438		;let BIOS have vblank
.regular
	move.w	#4,REG_IRQACK
	tst.b	vblankflag	;don't do anything if main routine hasn't finished yet
	beq	.return
.process
	movem.l	D0-D7/A0-A6,-(A7)		
	jsr	UpdateVRAM	;make VRAM changes based on commands in list
	jsr	BIOSF_VBL	;read controller etc.
	movem.l	(A7)+,D0-D7/A0-A6
	clr.b	vblankflag	;signal that vblank is processed
.return
	rte

;---

;updates VRAM using a list of commands

UpdateVRAM:
	movem.l	D0/A0/A1,-(SP)
	lea	vramlist,A0	;A0: incremental pointer to vramlist
	move.l	vramptr,A1	;A1: end of the list, used to determine when end of list is reached
.loopOuter
	cmp.l	A0,A1		;is this the  end of the list?
	beq	.return
.validEntry
	move.w	(A0)+,D0	;word 0: number of words in this entry
	move.w	(A0)+,-2(A6)	;word 1: VRAM addr
	move.w	(A0)+,2(A6)	;word 2: VRAM mod
	subq.w	#1,D0
.loopInner
	move.w	(A0)+,(A6)	;upload word
	dbra	D0,.loopInner
	bra	.loopOuter
.return
	move.l	#vramlist,vramptr	;reset pointer into vramlist for next frame
	movem.l	(SP)+,D0/A0/A1
	rts